==> pvfs-kernel-1.5.8-linux-2.4.tgz: pvfs-kernel-1.5.8-linux-2.4/libkpvfs/klib/mgrcomm.c <==
/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */
#define __NO_VERSION_

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#include "wrapper.h"
#include "lib.h"
#include "sockio.h"
#include "pvfs_config.h"

#ifndef PVFS_NR_OPEN
#define PVFS_NR_OPEN (NR_OPEN)
#endif

fdesc_p pfds[PVFS_NR_OPEN] = {NULL};

/* MGRINFO structure - this is used to keep up with managers in use
 *
 * DESCRIPTION:
 *
 * The idea here is to try to keep a single connection open to each
 * manager that we are communicating with.  To do this, we're going to
 * keep a little table and do our matching by comparing the port and 
 * address that we would use to connect to the manager.  We're going
 * to hold a maximum of MAXMGRS connections around.  There is no reason
 * why this number can't be bigger.
 *
 * NOTES ON FORKING PROCESSES:
 *
 * Experience now shows me that while this does indeed speed up metadata
 * operations quite a bit, it also causes some problems with programs
 * that fork children to do PVFS I/O (when they have already talked to
 * the manager and established a connection).
 *
 * Keep in mind that this situation causes any number of other potential
 * problems as well, particularly when a PVFS file is opened before the
 * fork occurs.  This kind if situation could possibly be handled via
 * the use of semaphores around all I/O operations, but that isn't
 * implemented and isn't going to be in this version of PVFS.  The next
 * version should handle that sort of thing in addition to being thread
 * safe.
 * 
 * SOLUTION TO FORKING PROCESSES:
 *
 * For now I'm going to add a "pid" field to the manager structure.
 * This will keep up with the pid of the owner of the connection.  If
 * the pid in the structure doesn't match the pid of the process
 * attempting to use the connection, we will close the connection,
 * initialize our structure, and open a new connection.
 *
 * This should allow programs such as Bonnie, which do not pass on PVFS
 * FDs but do talk to the manager before spawning processes, to operate
 * correctly.  It will not help situations where PVFS FDs are passed on.
 * 
 */
#define MAXMGRS 5
static int mgrcnt = 0;
static struct mgrinfoitem {
	unsigned short int port;
	unsigned int addr;
	int fd;
} mgrinfo[MAXMGRS];

/* PROTOTYPES */
static int addmgrinfo(struct sockaddr *saddr_p, int fd);
static int getmgrfd(struct sockaddr *saddr_p);
static int badmgrfd(int fd);

/* SEND_MREQ() - send a request to the manager and receive an ack
 *
 * This function handles sending a request to the manager, including
 * opening a connection to the manager if one is not already
 * established.  If an error occurs during the process, the socket will
 * be closed.  If the acknowledgement from the manager indicates
 * success, an open socket will be returned to the calling process so
 * that the process can receive trailing data (if any) or keep the
 * socket open.
 *
 * In the event that communication fails on a previously opened socket,
 * the code will attempt to establish a new connection.  It will only retry
 * in this manner once.
 *
 * Note: CURRENTLY NOT USED IN PVFS_OPEN()
 *
 * PARAMETERS:
 * dir_p  - directory metadata
 * req_p  - pointer to request structure
 * data_p - pointer to data to send following request (may be NULL)
 * ack_p  - pointer to memory location in which to store ack from mgr
 */

int send_mreq_saddr(struct sockaddr *saddr_p, mreq_p req_p, void *data_p, mack_p ack_p)
{
	int fd, err;
	char newconn=0;
	struct iovec vec[2];

	if (!req_p || !ack_p || !saddr_p) {
		return -EINVAL;
	}
make_new_conn:
	if ((fd = getmgrfd(saddr_p)) < 0) {
		if ((fd = new_sock()) < 0) {
				return(fd);
		}
		LOG1("Connecting to mgr with timeout afresh on %d\n",fd);

		if ((err = connect_timeout(fd, saddr_p, sizeof(struct sockaddr),MGRCOMM_CONNECT_TIMEOUT_SECS)) < 0) {
			Close(fd);
			return err;
		}
		newconn=1;
	}
	LOG1("Connected to mgr! on fd = %d \n",fd);
	/* send request */
	req_p->majik_nr = MGR_MAJIK_NR;
	req_p->release_nr = PVFS_RELEASE_NR;
	if (req_p->dsize > 0 && data_p) {
		/* send req and data with a single call */
		vec[0].iov_base = req_p;
		vec[0].iov_len = sizeof(mreq);
		vec[1].iov_base = data_p;
		vec[1].iov_len = req_p->dsize;
		LOG("Sending req+trailer to mgr! \n");
		if ((err = bsendv(fd, vec, 2)) < (int)(sizeof(mreq) + req_p->dsize)) {
			ERR1("send_mreq_saddr (sending req with bsendv) %d",err);
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}
	else /* no data -- stick to the bsend */ {
		LOG("Sending req only to mgr! \n");
		if ((err = bsend(fd, req_p, sizeof(mreq))) < (int) sizeof(mreq)) {
			PERROR("send_mreq_saddr (sending req)");
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}

	/* receive ack.  Give up if it does not happen in reasonable amount
	 * of time.
	 */
	LOG("Waiting for ack with timeout! \n");
	if((err = brecv_timeout(fd, ack_p, sizeof(mack), MGRCOMM_BRECV_TIMEOUT_SECS)) < (int)sizeof(mack)){
		ERR1("send_mreq_saddr (receiving ack) %d",err);
		Close(fd);
		if(!newconn){
			badmgrfd(fd);
			goto make_new_conn; /* try to reopen connection */
		}
		return(err);
	}
	LOG("Got ack! \n");
	if (ack_p->majik_nr != MGR_MAJIK_NR) /* serious error -- nonsense ack */ {
		Close(fd);
		if (!newconn) badmgrfd(fd);
		return(-EIO);
	}
	if (ack_p->release_nr != PVFS_RELEASE_NR) /* serious error -- releases don't match */
	{
		Close(fd);
		if (!newconn) badmgrfd(fd);
		return(-EIO);
	}
	/* not going to close connection, must save the state! */
	if (newconn) {
		err = addmgrinfo(saddr_p, fd);
		if (err != 0) {
			return -ENOMEM;
		}
	}

	if (ack_p->status) /* status error -- don't close the connection...  */ {
		return(-ack_p->eno);
	}
	return(fd);
} /* end of send_mreq_saddr() */

static int addmgrinfo(struct sockaddr *saddr_p, int fd)
{
	if (mgrcnt >= MAXMGRS) {
			ERR("addmgrinfo: MAXMGRS exceeded!\n");
			return(-EINVAL);
	}
	mgrinfo[mgrcnt].addr = ((struct sockaddr_in *) saddr_p)->sin_addr.s_addr;
	mgrinfo[mgrcnt].port = ((struct sockaddr_in *) saddr_p)->sin_port;
	mgrinfo[mgrcnt].fd = fd;
	mgrcnt++;

	/* update fd table */
	if (pfds[fd]) {
		LOG1("fd %d already in use?!?\n", fd);
	}
	pfds[fd] = (fdesc_p)kmalloc(sizeof(fdesc),GFP_KERNEL);
	if(!pfds[fd]) {
			LOG("Could not allocate memory\n");
			return -ENOMEM;
	}
	pfds[fd]->fs = FS_RESV;
	pfds[fd]->part_p = NULL;
	pfds[fd]->dim_p  = NULL;
	pfds[fd]->fd.ref = -1;
	pfds[fd]->fd.grp_nr = -1;
	return(0);
}

static int getmgrfd(struct sockaddr *saddr_p)
{
	int i;
	for (i = 0; i < mgrcnt; i++) {
		if (((struct sockaddr_in *)saddr_p)->sin_addr.s_addr == mgrinfo[i].addr
		&& ((struct sockaddr_in *)saddr_p)->sin_port == mgrinfo[i].port)
		{
			return(mgrinfo[i].fd);
		}
	}
	return(-1);
}

int init_mgrconn(void)
{
	int i;
	for (i = 0; i < mgrcnt; i++) {
		mgrinfo[i].fd = -1;
	}
	mgrcnt = 0;
	return(0);
}

void cleanup_mgrconn(void)
{
	int i;
	for(i=0;i < mgrcnt; i++) {
		if(mgrinfo[i].fd >= 0) {
			Close(mgrinfo[i].fd);
		}
		mgrinfo[i].fd = -1;
	}
	mgrcnt = 0;
	return;
}

static int badmgrfd(int fd)
{
	int i=0, found=-1;

	for (; i < mgrcnt; i++) {
		if (fd == mgrinfo[i].fd) found=i;
	}
	if (found < 0) return(-1);
	if (found != mgrcnt-1) {
		mgrinfo[found] = mgrinfo[mgrcnt-1];
	}
	mgrcnt--;

	/* update fd table */
	if (!pfds[fd] || pfds[fd]->fs != FS_RESV) {
		ERR1("fd %d was not reserved?!?\n", fd);
		return(-1);
	}
	kfree(pfds[fd]);
	pfds[fd] = NULL;
	return(0);
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 *
 * vim: ts=3
 */

==> pvfs-kernel-1.6.2-linux-2.4.tgz: pvfs-kernel-1.6.2-linux-2.4/libkpvfs/klib/mgrcomm.c <==
/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */
#define __NO_VERSION_

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#include "wrapper.h"
#include "lib.h"
#include "sockio.h"
#include "pvfs_config.h"


#ifndef PVFS_NR_OPEN
#define PVFS_NR_OPEN (NR_OPEN)
#endif

fdesc_p pfds[PVFS_NR_OPEN] = {NULL};

/* MGRINFO structure - this is used to keep up with managers in use
 *
 * DESCRIPTION:
 *
 * The idea here is to try to keep a single connection open to each
 * manager that we are communicating with.  To do this, we're going to
 * keep a little table and do our matching by comparing the port and 
 * address that we would use to connect to the manager.  We're going
 * to hold a maximum of MAXMGRS connections around.  There is no reason
 * why this number can't be bigger.
 *
 * NOTES ON FORKING PROCESSES:
 *
 * Experience now shows me that while this does indeed speed up metadata
 * operations quite a bit, it also causes some problems with programs
 * that fork children to do PVFS I/O (when they have already talked to
 * the manager and established a connection).
 *
 * Keep in mind that this situation causes any number of other potential
 * problems as well, particularly when a PVFS file is opened before the
 * fork occurs.  This kind if situation could possibly be handled via
 * the use of semaphores around all I/O operations, but that isn't
 * implemented and isn't going to be in this version of PVFS.  The next
 * version should handle that sort of thing in addition to being thread
 * safe.
 * 
 * SOLUTION TO FORKING PROCESSES:
 *
 * For now I'm going to add a "pid" field to the manager structure.
 * This will keep up with the pid of the owner of the connection.  If
 * the pid in the structure doesn't match the pid of the process
 * attempting to use the connection, we will close the connection,
 * initialize our structure, and open a new connection.
 *
 * This should allow programs such as Bonnie, which do not pass on PVFS
 * FDs but do talk to the manager before spawning processes, to operate
 * correctly.  It will not help situations where PVFS FDs are passed on.
 * 
 */
#define MAXMGRS 5
static int mgrcnt = 0;
static struct mgrinfoitem {
	unsigned short int port;
	unsigned int addr;
	int fd;
} mgrinfo[MAXMGRS];

/* PROTOTYPES */
static int addmgrinfo(struct sockaddr *saddr_p, int fd);
static int getmgrfd(struct sockaddr *saddr_p);
static int badmgrfd(int fd);

/* SEND_MREQ() - send a request to the manager and receive an ack
 *
 * This function handles sending a request to the manager, including
 * opening a connection to the manager if one is not already
 * established.  If an error occurs during the process, the socket will
 * be closed.  If the acknowledgement from the manager indicates
 * success, an open socket will be returned to the calling process so
 * that the process can receive trailing data (if any) or keep the
 * socket open.
 *
 * In the event that communication fails on a previously opened socket,
 * the code will attempt to establish a new connection.  It will only retry
 * in this manner once.
 *
 * Note: CURRENTLY NOT USED IN PVFS_OPEN()
 *
 * PARAMETERS:
 * dir_p  - directory metadata
 * req_p  - pointer to request structure
 * data_p - pointer to data to send following request (may be NULL)
 * ack_p  - pointer to memory location in which to store ack from mgr
 */

int send_mreq_saddr(struct sockaddr *saddr_p, mreq_p req_p, void *data_p, mack_p ack_p, void *recv_p)
{
	int fd, err;
	char newconn=0;
	struct iovec vec[2];
	int64_t ackdsize = 0;

	if (!req_p || !ack_p || !saddr_p) {
		return -EINVAL;
	}
make_new_conn:
	if ((fd = getmgrfd(saddr_p)) < 0) {
		if ((fd = new_sock()) < 0) {
				return(fd);
		}
		LOG1("Connecting to mgr with timeout afresh on %d\n",fd);

		if ((err = connect_timeout(fd, saddr_p, sizeof(struct sockaddr),MGRCOMM_CONNECT_TIMEOUT_SECS)) < 0) {
			Close(fd);
			return err;
		}
		newconn=1;
	}
	LOG1("Connected to mgr! on fd = %d \n",fd);
	/* send request */
	req_p->majik_nr = MGR_MAJIK_NR;
	req_p->release_nr = PVFS_RELEASE_NR;
	if (req_p->dsize > 0 && data_p) {
		/* send req and data with a single call */
		vec[0].iov_base = req_p;
		vec[0].iov_len = sizeof(mreq);
		vec[1].iov_base = data_p;
		vec[1].iov_len = req_p->dsize;
		LOG("Sending req+trailer to mgr! \n");
		if ((err = bsendv(fd, vec, 2)) < (int)(sizeof(mreq) + req_p->dsize)) {
			ERR1("send_mreq_saddr (sending req with bsendv) %d",err);
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}
	else /* no data -- stick to the bsend */ {
		LOG("Sending req only to mgr! \n");
		if ((err = bsend(fd, req_p, sizeof(mreq))) < (int) sizeof(mreq)) {
			PERROR("send_mreq_saddr (sending req)");
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}

	/* Do things differently if it is a open request with ackdsize != 0 */
	if(req_p->type == MGR_OPEN
			&& (ackdsize = req_p->req.open.ackdsize) != 0) {
		if(recv_p == NULL) {
			PERROR("send_mreq_saddr (receiving ack with bvrecv)");
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return -EFAULT;
		}
		/* recv ack and trailing data with a single call */
		vec[0].iov_base = ack_p;
		vec[0].iov_len = sizeof(mack);
		vec[1].iov_base = recv_p;
		vec[1].iov_len = ackdsize;
		if((err=bvrecv(fd, vec, 2)) < (int)(sizeof(mack) + ackdsize)) {
			PERROR("send_mreq_saddr (receiving ack with bvrecv)");
			Close(fd);
			if (!newconn) {
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}
	else {
		/* receive ack.  Give up if it does not happen in reasonable amount
		 * of time.
		 */
		LOG("Waiting for ack with timeout! \n");
		if((err = brecv_timeout(fd, ack_p, sizeof(mack), MGRCOMM_BRECV_TIMEOUT_SECS)) < (int)sizeof(mack)){
			ERR1("send_mreq_saddr (receiving ack) %d",err);
			Close(fd);
			if(!newconn){
				badmgrfd(fd);
				goto make_new_conn; /* try to reopen connection */
			}
			return(err);
		}
	}
	if (ack_p->majik_nr != MGR_MAJIK_NR) /* serious error -- nonsense ack */ {
		Close(fd);
		if (!newconn) badmgrfd(fd);
		return(-EIO);
	}
	if (ack_p->release_nr != PVFS_RELEASE_NR) /* serious error -- releases don't match */
	{
		Close(fd);
		if (!newconn) badmgrfd(fd);
		return(-EIO);
	}
	/* not going to close connection, must save the state! */
	if (newconn) {
		err = addmgrinfo(saddr_p, fd);
		if (err != 0) {
			return -ENOMEM;
		}
	}

	if (ack_p->status) /* status error -- don't close the connection...  */ {
		return(-ack_p->eno);
	}
	return(fd);
} /* end of send_mreq_saddr() */

static int addmgrinfo(struct sockaddr *saddr_p, int fd)
{
	if (mgrcnt >= MAXMGRS) {
			ERR("addmgrinfo: MAXMGRS exceeded!\n");
			return(-EINVAL);
	}
	mgrinfo[mgrcnt].addr = ((struct sockaddr_in *) saddr_p)->sin_addr.s_addr;
	mgrinfo[mgrcnt].port = ((struct sockaddr_in *) saddr_p)->sin_port;
	mgrinfo[mgrcnt].fd = fd;
	mgrcnt++;

	/* update fd table */
	if (pfds[fd]) {
		LOG1("fd %d already in use?!?\n", fd);
	}
	pfds[fd] = (fdesc_p)kmalloc(sizeof(fdesc),GFP_KERNEL);
	if(!pfds[fd]) {
			LOG("Could not allocate memory\n");
			return -ENOMEM;
	}
	pfds[fd]->fs = FS_RESV;
	pfds[fd]->part_p = NULL;
	pfds[fd]->dim_p  = NULL;
	pfds[fd]->fd.ref = -1;
	pfds[fd]->fd.grp_nr = -1;
	return(0);
}

static int getmgrfd(struct sockaddr *saddr_p)
{
	int i;
	for (i = 0; i < mgrcnt; i++) {
		if (((struct sockaddr_in *)saddr_p)->sin_addr.s_addr == mgrinfo[i].addr
		&& ((struct sockaddr_in *)saddr_p)->sin_port == mgrinfo[i].port)
		{
			return(mgrinfo[i].fd);
		}
	}
	return(-1);
}

int init_mgrconn(void)
{
	int i;
	for (i = 0; i < mgrcnt; i++) {
		mgrinfo[i].fd = -1;
	}
	mgrcnt = 0;
	return(0);
}

void cleanup_mgrconn(void)
{
	int i;
	for(i=0;i < mgrcnt; i++) {
		if(mgrinfo[i].fd >= 0) {
			Close(mgrinfo[i].fd);
		}
		mgrinfo[i].fd = -1;
	}
	mgrcnt = 0;
	return;
}

static int badmgrfd(int fd)
{
	int i=0, found=-1;

	for (; i < mgrcnt; i++) {
		if (fd == mgrinfo[i].fd) found=i;
	}
	if (found < 0) return(-1);
	if (found != mgrcnt-1) {
		mgrinfo[found] = mgrinfo[mgrcnt-1];
	}
	mgrcnt--;

	/* update fd table */
	if (!pfds[fd] || pfds[fd]->fs != FS_RESV) {
		ERR1("fd %d was not reserved?!?\n", fd);
		return(-1);
	}
	kfree(pfds[fd]);
	pfds[fd] = NULL;
	return(0);
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 *
 * vim: ts=3
 */
